/*
 *  Gazebo - Outdoor Multi-Robot Simulator
 *  Copyright (C) 2003
 *     Nate Koenig & Andrew Howard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Desc: Bumper controller
 * Author: Nate Koenig
 * Date: 09 Setp. 2008
 */

#include <gazebo_plugins/gazebo_ros_bumper.h>

#include <gazebo/Global.hh>
#include <gazebo/XMLConfig.hh>
#include <gazebo/ContactSensor.hh>
#include <gazebo/World.hh>
#include <gazebo/gazebo.h>
#include <gazebo/GazeboError.hh>
#include <gazebo/ControllerFactory.hh>
#include <gazebo/Simulator.hh>
#include <gazebo/Body.hh>
#include <gazebo/Model.hh>
#include <gazebo/Geom.hh>

namespace gazebo
{

GZ_REGISTER_DYNAMIC_CONTROLLER("gazebo_ros_bumper", GazeboRosBumper);

////////////////////////////////////////////////////////////////////////////////
// Constructor
GazeboRosBumper::GazeboRosBumper(Entity *parent )
  : Controller(parent)
{
  this->myParent = dynamic_cast<ContactSensor*>(this->parent);

  if (!this->myParent)
    gzthrow("Bumper controller requires a Contact Sensor as its parent");

  Param::Begin(&this->parameters);
  this->bumperTopicNameP = new ParamT<std::string>("bumperTopicName", "", 1);
  this->robotNamespaceP = new ParamT<std::string>("robotNamespace", "/", 0);
  Param::End();

  this->contactConnectCount = 0;
}

////////////////////////////////////////////////////////////////////////////////
// Destructor
GazeboRosBumper::~GazeboRosBumper()
{
  delete this->robotNamespaceP;
  delete this->bumperTopicNameP;
  delete this->rosnode_;
}

////////////////////////////////////////////////////////////////////////////////
// Load the controller
void GazeboRosBumper::LoadChild(XMLConfigNode *node)
{
  this->robotNamespaceP->Load(node);
  this->robotNamespace = this->robotNamespaceP->GetValue();

  if (!ros::isInitialized())
  {
    int argc = 0;
    char** argv = NULL;
    ros::init(argc,argv,"gazebo",ros::init_options::NoSigintHandler|ros::init_options::AnonymousName);
  }
  this->rosnode_ = new ros::NodeHandle(this->robotNamespace);

  this->bumperTopicNameP->Load(node);
  this->bumperTopicName = this->bumperTopicNameP->GetValue();
  ROS_DEBUG("publishing contact/collisions to topic name: %s", 
           this->bumperTopicName.c_str());
  //std::cout << " publishing contact/collisions to topic name: " << this->bumperTopicName << std::endl;

#ifdef USE_CBQ
  ros::AdvertiseOptions ao = ros::AdvertiseOptions::create<gazebo_plugins::ContactsState>(
    this->bumperTopicName+std::string("/state"),1,
    boost::bind( &GazeboRosBumper::ContactConnect,this),
    boost::bind( &GazeboRosBumper::ContactDisconnect,this), ros::VoidPtr(), &this->contact_queue_);
  this->contact_pub_ = this->rosnode_->advertise(ao);
#else
  this->contact_pub_ = this->rosnode_->advertise<gazebo_plugins::ContactsState>(this->bumperTopicName+std::string("/state"),1);
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Increment count
void GazeboRosBumper::ContactConnect()
{
  this->contactConnectCount++;
}
////////////////////////////////////////////////////////////////////////////////
// Decrement count
void GazeboRosBumper::ContactDisconnect()
{
  this->contactConnectCount--;
}

////////////////////////////////////////////////////////////////////////////////
// Initialize the controller
void GazeboRosBumper::InitChild()
{
#ifdef USE_CBQ
  // start custom queue for contact bumper
  this->contact_callback_queue_thread_ = new boost::thread( boost::bind( &GazeboRosBumper::ContactQueueThread,this ) );
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Update the controller
void GazeboRosBumper::UpdateChild()
{
  if (this->contactConnectCount > 0)
  {
    boost::mutex::scoped_lock sclock(this->lock);

    unsigned int num_contact_count = this->myParent->GetContactCount();

    std::string geom1_name;
    int in_contact;
    double contact_time;
    std::string geom2_name;

    // populate header
    //if (cur_body)
    //  this->contactMsg.header.frame_id = cur_body->GetName();  // @todo: transform results to the link name
    double cur_time = Simulator::Instance()->GetSimTime();
    this->contactsStateMsg.header.frame_id = "/map";  // information are in inertial coordinates
    this->contactsStateMsg.header.stamp.fromSec(cur_time);

    // set contact states size
    this->contactsStateMsg.set_states_size(num_contact_count);  // one contact_count per pair of geoms in contact

    // get contact array sizes for this contact param.  each contact param has num_contact_count contact pairs,
    //std::vector<Geom*> contact_geoms;
    int total_contact_points = 0;
    for (unsigned int i=0; i < num_contact_count; i++)
    {
      geom1_name   = this->myParent->GetGeomName(i); 
      geom2_name   = this->myParent->GetContactGeomName(i); 
      in_contact   = this->myParent->GetContactState(i); 
      contact_time = this->myParent->GetContactTime(i); 

      if (in_contact == 1)
      {
        // get pointer to Geom by name
        Model *model = Simulator::Instance()->GetParentModel(this->myParent);
        Geom *geom1 = model->GetGeom(geom1_name);
        if (geom1)
        {
          // add to list of contact_geoms
          //contact_geoms.push_back(geom1);
          std::ostringstream stream;
          stream    << "touched!    i:" << i
                    << "      my geom:" << geom1_name
                    << "   other geom:" << geom2_name
                    << "         time:" << contact_time
                    << std::endl;
          //std::cout << stream.str();
          this->contactsStateMsg.states[i].info = stream.str();
          this->contactsStateMsg.states[i].geom1_name = geom1_name;
          this->contactsStateMsg.states[i].geom2_name = geom2_name;

          // make sure we have less contact points than preset buffer size
          int pts = geom1->contactPoints;
          if (pts > 0)
          {
            // set message array sizes, @todo: is there memory leak? do I need to reset each array?
            this->contactsStateMsg.states[i].set_wrenches_size(pts);
            this->contactsStateMsg.states[i].set_contact_positions_size(pts);
            this->contactsStateMsg.states[i].set_contact_normals_size(pts);
            this->contactsStateMsg.states[i].set_depths_size(pts);

            for (int j=0; j < pts; j++)
            {
              // set wrenches
              dJointFeedback contact_forces = this->myParent->GetContactFeedback(i,j); 
              this->contactsStateMsg.states[i].wrenches[j].force.x  = contact_forces.f1[0];
              this->contactsStateMsg.states[i].wrenches[j].force.y  = contact_forces.f1[1];
              this->contactsStateMsg.states[i].wrenches[j].force.z  = contact_forces.f1[2];
              this->contactsStateMsg.states[i].wrenches[j].torque.x = contact_forces.t1[0];
              this->contactsStateMsg.states[i].wrenches[j].torque.y = contact_forces.t1[1];
              this->contactsStateMsg.states[i].wrenches[j].torque.z = contact_forces.t1[2];

              // set contact positions
              this->contactsStateMsg.states[i].contact_positions[j].x = geom1->contactGeoms[j].pos[0];
              this->contactsStateMsg.states[i].contact_positions[j].y = geom1->contactGeoms[j].pos[1];
              this->contactsStateMsg.states[i].contact_positions[j].z = geom1->contactGeoms[j].pos[2];
              // set contact normals
              this->contactsStateMsg.states[i].contact_normals[j].x = geom1->contactGeoms[j].normal[0];
              this->contactsStateMsg.states[i].contact_normals[j].y = geom1->contactGeoms[j].normal[1];
              this->contactsStateMsg.states[i].contact_normals[j].z = geom1->contactGeoms[j].normal[2];
              // set contact depth, interpenetration
              this->contactsStateMsg.states[i].depths[j]            = geom1->contactGeoms[j].depth;
            }

            // keep track of total contact points
            total_contact_points += pts;
          }
          geom1->contactPoints = 0; // reset to 0 @todo: find a more elegant way
        }
      }
    }
    if (total_contact_points == 0)
    {
      this->contactsStateMsg.set_states_size(0);
    }
    else /* publish when there's contact */
      this->contact_pub_.publish(this->contactsStateMsg);

    this->myParent->ResetContactStates();

  }

}

////////////////////////////////////////////////////////////////////////////////
// Finalize the controller
void GazeboRosBumper::FiniChild()
{
  this->rosnode_->shutdown();
}

#ifdef USE_CBQ
////////////////////////////////////////////////////////////////////////////////
// Put laser data to the interface
void GazeboRosBumper::ContactQueueThread()
{
  static const double timeout = 0.01;

  while (this->rosnode_->ok())
  {
    this->contact_queue_.callAvailable(ros::WallDuration(timeout));
  }
}
#endif

}
